"""Tests of Braze task code."""

import logging
from unittest import TestCase
from unittest.mock import patch, MagicMock, Mock
from urllib.parse import urljoin

import braze.exceptions as edx_braze_exceptions
import ddt
import responses
from celery.exceptions import Retry
from requests.exceptions import HTTPError
from testfixtures import LogCapture

from ecommerce_worker.email.v1.braze.exceptions import (
    BrazeError,
    BrazeInternalServerError,
    BrazeRateLimitError
)
from ecommerce_worker.email.v1.tasks import (
    send_code_assignment_nudge_email,
    send_offer_assignment_email,
    send_offer_usage_email,
    send_offer_update_email,
    send_api_triggered_offer_usage_email,
)
from ecommerce_worker.utils import get_configuration


LOG_NAME = 'ecommerce_worker.email.v1.braze.tasks'
LOG_TASK_NAME = 'ecommerce_worker.email.v1.braze.tasks'
USER_EMAIL = 'user@unknown.com'
EMAILS = 'user@unknown.com, user1@example.com'
LMS_USER_IDS_BY_EMAIL = {
    'user@unknown.com': None,
    'user1@example.com': 456,
}
OFFER_ASSIGNMENT_ID = '555'
SEND_ID = '66cdc28f8f082bc3074c0c79f'
SUBJECT = 'New edX course assignment'
EMAIL_BODY = 'Template message with johndoe@unknown.com GIL7RUEOU7VHBH7Q ' \
             'https://tempurl.url/enroll 3 2012-04-23'
USAGE_EMAIL_BODY_VARIABLES = {
    'FOO': 123,
    'BAR': 555,
}
SENDER_ALIAS = 'edx Support Team'
REPLY_TO = 'reply@edx.org'
SITE_CODE = 'test'
BRAZE_CONFIG = {
    'BRAZE_ENABLE': True,
    'BRAZE_REST_API_KEY': 'rest_api_key',
    'BRAZE_WEBAPP_API_KEY': 'webapp_api_key',
    'REST_API_URL': 'https://rest.iad-06.braze.com',
    'MESSAGES_SEND_ENDPOINT': '/messages/send',
    'EMAIL_BOUNCE_ENDPOINT': '/email/hard_bounces',
    'NEW_ALIAS_ENDPOINT': '/users/alias/new',
    'USERS_TRACK_ENDPOINT': '/users/track',
    'EXPORT_ID_ENDPOINT': '/users/export/ids',
    'CAMPAIGN_SEND_ENDPOINT': '/campaigns/trigger/send',
    'ENTERPRISE_CAMPAIGN_ID': '',
    'ENTERPRISE_CODE_ASSIGNMENT_CAMPAIGN_ID': 'the-code-assignment-campaign-id',
    'ENTERPRISE_CODE_UPDATE_CAMPAIGN_ID': 'the-code-update-campaign-id',
    'ENTERPRISE_CODE_USAGE_API_TRIGGERED_CAMPAIGN_ID': 'the-code-usage-api-triggerd-campaign-id',
    'ENTERPRISE_CODE_USAGE_CAMPAIGN_ID': None,
    'ENTERPRISE_CODE_NUDGE_CAMPAIGN_ID': '',  # cover case where a `campaign_id` kwarg is falsey.
    'FROM_EMAIL': '<edx-for-business-no-reply@info.edx.org>',
    'BRAZE_RETRY_SECONDS': 3600,
    'BRAZE_RETRY_ATTEMPTS': 6,
}


@ddt.ddt
class SendEmailsViaBrazeTests(TestCase):
    """ Validates the email sending tasks with Braze api. """
    ASSIGNMENT_TASK_KWARGS = {
        'user_email': USER_EMAIL,
        'offer_assignment_id': OFFER_ASSIGNMENT_ID,
        'subject': SUBJECT,
        'email_body': EMAIL_BODY,
        'sender_alias': SENDER_ALIAS,
        'reply_to': REPLY_TO,
        'site_code': SITE_CODE,
    }

    UPDATE_TASK_KWARGS = {
        'user_email': USER_EMAIL,
        'subject': SUBJECT,
        'email_body': EMAIL_BODY,
        'sender_alias': SENDER_ALIAS,
        'reply_to': REPLY_TO,
        'site_code': SITE_CODE,
    }

    USAGE_TASK_KWARGS = {
        'emails': EMAILS,
        'subject': SUBJECT,
        'email_body': EMAIL_BODY,
        'reply_to': REPLY_TO,
        'site_code': SITE_CODE,
    }

    NUDGE_TASK_KWARGS = {
        'email': USER_EMAIL,
        'subject': SUBJECT,
        'email_body': EMAIL_BODY,
        'sender_alias': SENDER_ALIAS,
        'reply_to': REPLY_TO,
        'site_code': SITE_CODE,
    }

    OAUTH_ACCESS_TOKEN_URL = urljoin(
        get_configuration('BACKEND_SERVICE_EDX_OAUTH2_PROVIDER_URL') + '/', 'access_token/'
    )

    def execute_task(self):
        """ Execute the send_offer_assignment_email task. """
        with patch('ecommerce_worker.configuration.test.BRAZE', BRAZE_CONFIG):
            send_offer_assignment_email(**self.ASSIGNMENT_TASK_KWARGS)

    def mock_braze_user_endpoints(self):
        """ Mock POST requests to the user alias and track endpoints. """
        responses.add(
            responses.POST,
            'https://rest.iad-06.braze.com/users/track',
            json={'message': 'success'},
            status=201
        )
        responses.add(
            responses.POST,
            'https://rest.iad-06.braze.com/users/alias/new',
            json={'message': 'success'},
            status=201
        )
        responses.add(
            responses.POST,
            'https://rest.iad-06.braze.com/users/export/ids',
            json={"users": [], "message": "success"},
            status=201
        )

    @patch('ecommerce_worker.email.v1.braze.tasks.get_braze_client', Mock(side_effect=BrazeError))
    @ddt.data(
        (send_offer_assignment_email, ASSIGNMENT_TASK_KWARGS, "Offer Assignment", 'assignment'),
        (send_offer_update_email, UPDATE_TASK_KWARGS, "Offer Assignment", 'update'),
        (send_offer_usage_email, USAGE_TASK_KWARGS, "Offer Usage", 'usage'),
        (send_code_assignment_nudge_email, NUDGE_TASK_KWARGS, "Code Assignment Nudge Email", 'nudge'),
    )
    @ddt.unpack
    def test_client_instantiation_error(self, task, task_kwargs, logger_prefix, log_message):
        """ Verify no message is sent if an error occurs while instantiating the Braze API client. """
        with patch('ecommerce_worker.configuration.test.BRAZE', BRAZE_CONFIG):
            with LogCapture(level=logging.INFO) as log:
                task(**task_kwargs)
        log.check(
            (
                LOG_NAME,
                'ERROR',
                '[{logger_prefix}] Error in offer {log_message} notification with message --- '
                '{message}'.format(
                    logger_prefix=logger_prefix,
                    log_message=log_message,
                    message=EMAIL_BODY
                )
            ),
        )

    @patch('ecommerce_worker.email.v1.braze.tasks.logger.exception')
    @ddt.data(
        (send_offer_assignment_email, ASSIGNMENT_TASK_KWARGS, "Offer Assignment", 'assignment'),
        (send_offer_update_email, UPDATE_TASK_KWARGS, "Offer Assignment", 'update'),
        (send_offer_usage_email, USAGE_TASK_KWARGS, "Offer Usage", 'usage'),
        (send_code_assignment_nudge_email, NUDGE_TASK_KWARGS, "Code Assignment Nudge Email", 'nudge'),
    )
    @ddt.unpack
    def test_api_client_error(self, task, task_kwargs, logger_prefix, log_message, mock_log):
        """ Verify API client errors are logged. """
        with patch('ecommerce_worker.configuration.test.BRAZE', BRAZE_CONFIG):
            task(**task_kwargs)
        mock_log.assert_called_once_with(
            '[{logger_prefix}] Error in offer {log_message} notification with message --- '
            '{message}'.format(
                logger_prefix=logger_prefix,
                log_message=log_message,
                message=EMAIL_BODY
            )
        )

    @responses.activate
    @ddt.data(
        (send_offer_assignment_email, ASSIGNMENT_TASK_KWARGS),
        (send_offer_update_email, UPDATE_TASK_KWARGS),
        (send_offer_usage_email, USAGE_TASK_KWARGS),
        (send_code_assignment_nudge_email, NUDGE_TASK_KWARGS),
    )
    @ddt.unpack
    def test_api_429_error_with_retry(self, task, task_kwargs):
        """ Verify the task is rescheduled if an API error occurs, and the request can be retried. """
        self.mock_braze_user_endpoints()
        failure_response = {
            'message': 'Not a Success',
            'status_code': 429
        }
        host = 'https://rest.iad-06.braze.com/messages/send'
        responses.add(
            responses.POST,
            host,
            json=failure_response,
            status=429
        )
        with patch('ecommerce_worker.configuration.test.BRAZE', BRAZE_CONFIG):
            with self.assertRaises((BrazeRateLimitError, Retry)):
                task(**task_kwargs)

    @responses.activate
    @ddt.data(
        (send_offer_assignment_email, ASSIGNMENT_TASK_KWARGS),
        (send_offer_update_email, UPDATE_TASK_KWARGS),
        (send_offer_usage_email, USAGE_TASK_KWARGS),
        (send_code_assignment_nudge_email, NUDGE_TASK_KWARGS),
    )
    @ddt.unpack
    def test_api_500_error_with_retry(self, task, task_kwargs):
        """ Verify 500 error triggers a request retry. """
        self.mock_braze_user_endpoints()
        failure_response = {
            'message': 'Not a Success',
            'status_code': 500
        }
        host = 'https://rest.iad-06.braze.com/messages/send'
        responses.add(
            responses.POST,
            host,
            json=failure_response,
            status=500
        )
        with patch('ecommerce_worker.configuration.test.BRAZE', BRAZE_CONFIG):
            with self.assertRaises((BrazeInternalServerError, Retry)):
                task(**task_kwargs)

    @responses.activate
    @ddt.data(
        (send_offer_update_email, UPDATE_TASK_KWARGS),
        (send_offer_usage_email, USAGE_TASK_KWARGS),
        (send_code_assignment_nudge_email, NUDGE_TASK_KWARGS),
    )
    @ddt.unpack
    def test_api(self, task, task_kwargs):
        """
        Test the happy path.
        """
        self.mock_braze_user_endpoints()
        success_response = {
            "dispatch_id": "66cdc28f8f082bc3074c0c79f",
            "errors": [],
            "message": "success",
            "status_code": 201
        }
        host = 'https://rest.iad-06.braze.com/messages/send'
        responses.add(
            responses.POST,
            host,
            json=success_response,
            status=201)
        with patch('ecommerce_worker.configuration.test.BRAZE', BRAZE_CONFIG):
            task(**task_kwargs)
        self.assertIn('success', responses.calls[0].response.text)

    @responses.activate
    @patch('ecommerce_worker.email.v1.braze.tasks.update_assignment_email_status')
    def test_message_sent(self, mock_update_assignment):
        """ Verify a message is logged after a successful API call to send the message. """
        self.mock_braze_user_endpoints()
        success_response = {
            'dispatch_id': '66cdc28f8f082bc3074c0c79f',
            'errors': [],
            'message': 'success',
            'status_code': 201
        }
        host = 'https://rest.iad-06.braze.com/messages/send'
        responses.add(
            responses.POST,
            host,
            json=success_response,
            status=201)
        with LogCapture(level=logging.INFO) as log:
            self.execute_task()
        log.check_present(
            (
                LOG_TASK_NAME,
                'INFO',
                '[Offer Assignment] Offer assignment notification sent with message --- '
                '{message}'.format(message=EMAIL_BODY)
            ),
        )
        self.assertEqual(mock_update_assignment.call_count, 1)

    @responses.activate
    @patch('ecommerce_worker.email.v1.utils.get_access_token')
    @patch('ecommerce_worker.email.v1.utils.requests.post')
    def test_update_assignment_exception(self, mock_access_token, mock_requests_post):
        """
        Verify a message is logged after an unsuccessful API call to update the status.
        """
        self.mock_braze_user_endpoints()
        success_response = {
            'dispatch_id': '66cdc28f8f082bc3074c0c79f',
            'errors': [],
            'message': 'success',
            'status_code': 201
        }
        host = 'https://rest.iad-06.braze.com/messages/send'
        responses.add(
            responses.POST,
            host,
            json=success_response,
            status=201
        )

        mock_access_token.return_value.json.return_value = {'access_token': 'FAKE-access-token'}
        mock_requests_post.return_value.raise_for_status.side_effect = HTTPError

        with LogCapture(level=logging.INFO) as log:
            self.execute_task()
        log.check_present(
            (
                LOG_TASK_NAME,
                'ERROR',
                '[Offer Assignment] An error occurred while updating email status data for offer {token_offer} and '
                'email {token_email} via the ecommerce API.'.format(
                    token_offer=OFFER_ASSIGNMENT_ID,
                    token_email=USER_EMAIL
                )
            )
        )

    def test_braze_not_enabled(self):
        """
        Test that tasks do nothing when Braze is not enabled.
        """
        # pylint: disable=unnecessary-comprehension
        disabled_config = {key: value for key, value in BRAZE_CONFIG.items()}
        disabled_config['BRAZE_ENABLE'] = False
        for task, braze_task_name, kwargs in (
            (send_code_assignment_nudge_email, 'send_code_assignment_nudge_email_via_braze', self.NUDGE_TASK_KWARGS),
            (send_offer_usage_email, 'send_offer_usage_email_via_braze', self.USAGE_TASK_KWARGS),
            (send_offer_assignment_email, 'send_offer_assignment_email_via_braze', self.ASSIGNMENT_TASK_KWARGS),
            (send_offer_update_email, 'send_offer_update_email_via_braze', self.UPDATE_TASK_KWARGS),
        ):
            task_path = 'ecommerce_worker.email.v1.tasks.{}'.format(braze_task_name)
            with patch(task_path) as mock_braze_task:
                task(**kwargs)
                self.assertFalse(mock_braze_task.called)


@ddt.ddt
class OfferUsageEmailTests(TestCase):
    """
    Separate test class for the offer usage email task, since it utilizes
    edx-braze-client.
    """
    USAGE_TASK_KWARGS = {
        'lms_user_ids_by_email': LMS_USER_IDS_BY_EMAIL,
        'subject': SUBJECT,
        'email_body_variables': USAGE_EMAIL_BODY_VARIABLES,
        'site_code': SITE_CODE,
    }

    def test_braze_not_enabled(self):
        """
        Test that this task does nothing when Braze is not enabled.
        """
        # pylint: disable=unnecessary-comprehension
        disabled_config = {key: value for key, value in BRAZE_CONFIG.items()}
        disabled_config['BRAZE_ENABLE'] = False
        task_path = 'ecommerce_worker.email.v1.tasks.send_api_triggered_offer_usage_email_via_braze'
        with patch(task_path) as mock_braze_task:
            send_api_triggered_offer_usage_email(**self.USAGE_TASK_KWARGS)
            self.assertFalse(mock_braze_task.called)

    @patch('ecommerce_worker.email.v1.braze.tasks.EdxBrazeClient', return_value=MagicMock())
    @ddt.data(
        edx_braze_exceptions.BrazeRateLimitError,
        edx_braze_exceptions.BrazeInternalServerError,
    )
    def test_api_error_with_retry(self, exception, mock_client):
        """ Verify the task is rescheduled if an expected API error occurs, and the request can be retried. """
        mock_client.return_value.send_campaign_message.side_effect = exception('msg')
        with patch('ecommerce_worker.configuration.test.BRAZE', BRAZE_CONFIG):
            with self.assertRaises((exception, Retry)):
                send_api_triggered_offer_usage_email(**self.USAGE_TASK_KWARGS)

    @patch('ecommerce_worker.email.v1.braze.tasks.EdxBrazeClient', return_value=MagicMock())
    def test_api_generic_braze_error(self, mock_client):
        """ Verify the task raises a BrazeError for other types of exceptions. """
        mock_client.return_value.send_campaign_message.side_effect = edx_braze_exceptions.BrazeError('msg')
        with patch('ecommerce_worker.configuration.test.BRAZE', BRAZE_CONFIG):
            with self.assertRaises(edx_braze_exceptions.BrazeError):
                send_api_triggered_offer_usage_email(**self.USAGE_TASK_KWARGS)

    @patch('ecommerce_worker.email.v1.braze.tasks.EdxBrazeClient', return_value=MagicMock(), autospec=True)
    @ddt.data(
        None,
        'custom-campaign-id',
    )
    def test_successful_usage_email(self, campaign_id, mock_client):
        """ Verify the task successfully calls send_campaign_message() and create_recipient(). """
        mock_client.reset_mock()

        def mock_create_recipient(user_email, lms_user_id):
            return {'user_email': user_email, 'lms_user_id': lms_user_id}

        client = mock_client.return_value
        client.create_recipient.side_effect = mock_create_recipient

        with patch('ecommerce_worker.configuration.test.BRAZE', BRAZE_CONFIG):
            send_api_triggered_offer_usage_email(campaign_id=campaign_id, **self.USAGE_TASK_KWARGS)

        client.create_recipient.assert_called_once_with(
            'user1@example.com', 456,
        )

        expected_trigger_properties = {
            'subject': 'New edX course assignment',
            **USAGE_EMAIL_BODY_VARIABLES,
        }
        expected_campaign_id = campaign_id or BRAZE_CONFIG['ENTERPRISE_CODE_USAGE_API_TRIGGERED_CAMPAIGN_ID']
        client.send_campaign_message.assert_called_once_with(
            campaign_id=expected_campaign_id,
            recipients=[
                {'user_email': 'user1@example.com', 'lms_user_id': 456}
            ],
            emails=['user@unknown.com'],
            trigger_properties=expected_trigger_properties,
        )



"""Logging configuration"""
from logging.handlers import SysLogHandler
import os
import platform
import sys


def get_logger_config(log_dir='/var/tmp',
                      logging_env='no_env',
                      edx_filename='edx.log',
                      dev_env=False,
                      debug=False,
                      local_loglevel='INFO',
                      service_variant='ecomworker'):

    """
    Returns a dictionary containing logging configuration.

    If dev_env is True, logging will not be done via local rsyslogd.
    Instead, application logs will be dropped into log_dir. 'edx_filename'
    is ignored unless dev_env is True.
    """
    # Revert to INFO if an invalid string is passed in
    if local_loglevel not in ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']:
        local_loglevel = 'INFO'

    hostname = platform.node().split('.')[0]
    syslog_format = (
        '[service_variant={service_variant}]'
        '[%(name)s][env:{logging_env}] %(levelname)s '
        '[{hostname}  %(process)d] [%(filename)s:%(lineno)d] '
        '- %(message)s'
    ).format(
        service_variant=service_variant,
        logging_env=logging_env, hostname=hostname
    )

    if debug:
        handlers = ['console']
    else:
        handlers = ['local']

    logger_config = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': '%(asctime)s %(levelname)s %(process)d '
                          '[%(name)s] %(filename)s:%(lineno)d - %(message)s',
            },
            'syslog_format': {'format': syslog_format},
            'raw': {'format': '%(message)s'},
        },
        'handlers': {
            'console': {
                'level': 'DEBUG' if debug else 'INFO',
                'class': 'logging.StreamHandler',
                'formatter': 'standard',
                'stream': sys.stdout,
            },
        },
        'loggers': {
            'requests': {
                'handlers': handlers,
                'level': 'WARNING',
                'propagate': True
            },
            '': {
                'handlers': handlers,
                'level': 'DEBUG',
                'propagate': False
            },
        }
    }

    if dev_env:
        edx_file_loc = os.path.join(log_dir, edx_filename)
        logger_config['handlers'].update({
            'local': {
                'class': 'logging.handlers.RotatingFileHandler',
                'level': local_loglevel,
                'formatter': 'standard',
                'filename': edx_file_loc,
                'maxBytes': 1024 * 1024 * 2,
                'backupCount': 5,
            },
        })
    else:
        logger_config['handlers'].update({
            'local': {
                'level': local_loglevel,
                'class': 'logging.handlers.SysLogHandler',
                # Use a different address for Mac OS X
                'address': '/var/run/syslog' if sys.platform == 'darwin' else '/dev/log',
                'formatter': 'syslog_format',
                'facility': SysLogHandler.LOG_LOCAL0,
            },
        })

    return logger_config



"""Tests of cache."""
import logging
from unittest import TestCase

from ecommerce_worker.cache import Cache

log = logging.getLogger(__name__)


class CacheTests(TestCase):
    """
    Tests for the cache routine.
    """

    def test_content_cache(self):
        """
        Test content cache code
        """

        # Create a cache and add 5 items to it
        cache = Cache()
        cache.set('key1', 'value1', 100)
        cache.set('key2', 'value2', 100)
        # set duration to zero or negative so they expire
        cache.set('key3', 'value3', 0)
        cache.set('key4', 'value4', -100)
        cache.set('key5', 'value5', -100)

        # should be 5 to start
        self.assertEqual(len(cache), 5)

        # getting one of the expired should clean out all expired
        self.assertEqual(cache.get('key5'), None)
        # make sure 2 left
        self.assertEqual(len(cache), 2)
        self.assertEqual(cache.get('key2'), 'value2')
        self.assertEqual(cache.get('key1'), 'value1')
        self.assertEqual(cache.get('key3'), None)



# pylint: disable=cyclic-import
"""
Braze Client functions.
"""

from urllib.parse import urlencode, urljoin

import copy
import json
import requests
from celery.utils.log import get_task_logger

from braze import client as edx_braze_client

from ecommerce_worker.email.v1.braze.exceptions import (
    ConfigurationError,
    BrazeNotEnabled,
    BrazeClientError,
    BrazeRateLimitError,
    BrazeInternalServerError
)
from ecommerce_worker.utils import get_configuration

log = get_task_logger(__name__)


def is_braze_enabled(site_code) -> bool:
    config = get_braze_configuration(site_code)
    return bool(config.get('BRAZE_ENABLE'))


def get_braze_configuration(site_code):
    """
    Returns the Braze configuration for the specified site.

    Arguments:
        site_code (str): Site for fetching the configuration.
    """
    config = get_configuration('BRAZE', site_code=site_code)
    return config


def validate_braze_config(config, site_code):
    """
    Raises if braze is not enabled or if either the
    Rest or Webapp API keys are missing from the configuration.
    """
    # Return if Braze integration disabled
    if not config.get('BRAZE_ENABLE'):
        msg = f'Braze is not enabled for site {site_code}'
        log.error(msg)
        raise BrazeNotEnabled(msg)

    if not (config.get('BRAZE_REST_API_KEY') and config.get('BRAZE_WEBAPP_API_KEY')):
        msg = f'Required keys missing for site {site_code}'
        log.error(msg)
        raise ConfigurationError(msg)


def get_braze_client(site_code):
    """
    Returns a Braze client for the specified site.

    Arguments:
        site_code (str): Site for which the client should be configured.

    Returns:
        BrazeClient

    Raises:
        BrazeNotEnabled: If Braze is not enabled for the specified site.
        ConfigurationError: If either the Braze API key or Webapp key are not set for the site.
    """
    config = get_braze_configuration(site_code)
    validate_braze_config(config, site_code)

    return BrazeClient(
        rest_api_key=config.get('BRAZE_REST_API_KEY'),
        webapp_api_key=config.get('BRAZE_WEBAPP_API_KEY'),
        rest_api_url=config.get('REST_API_URL'),
        messages_send_endpoint=config.get('MESSAGES_SEND_ENDPOINT'),
        email_bounce_endpoint=config.get('EMAIL_BOUNCE_ENDPOINT'),
        new_alias_endpoint=config.get('NEW_ALIAS_ENDPOINT'),
        users_track_endpoint=config.get('USERS_TRACK_ENDPOINT'),
        export_id_endpoint=config.get('EXPORT_ID_ENDPOINT'),
        campaign_send_endpoint=config.get('CAMPAIGN_SEND_ENDPOINT'),
        enterprise_campaign_id=config.get('ENTERPRISE_CAMPAIGN_ID'),
        from_email=config.get('FROM_EMAIL'),
    )


class BrazeClient:
    """
    Client for Braze REST API
    """

    def __init__(
            self,
            rest_api_key,
            webapp_api_key,
            rest_api_url,
            messages_send_endpoint,
            email_bounce_endpoint,
            new_alias_endpoint,
            users_track_endpoint,
            export_id_endpoint,
            campaign_send_endpoint,
            enterprise_campaign_id,
            from_email,
    ):
        """
        Initialize the Braze Client with configuration values.

        Arguments:
            rest_api_key (str): API key to use for accessing Braze REST endpoints
            webapp_api_key (str): Key that identifies the Braze webapp application
            rest_api_url    (str): REST API url,
            messages_send_endpoint  (str): Messages send endpoint,
            email_bounce_endpoint   (str): Email bounce endpoint,
            new_alias_endpoint  (str): New alias endpoint,
            users_track_endpoint    (str): Users track endpoint,
            from_email (str): Braze email from address
            export_id_endpoint (str): User export endpoint
            campaign_send_endpoint (str): Campaign send endpoint
            enterprise_campaign_id (str): Campaign identifier
        """
        self.rest_api_key = rest_api_key
        self.webapp_api_key = webapp_api_key
        self.rest_api_url = rest_api_url
        self.messages_send_endpoint = messages_send_endpoint
        self.email_bounce_endpoint = email_bounce_endpoint
        self.new_alias_endpoint = new_alias_endpoint
        self.users_track_endpoint = users_track_endpoint
        self.export_id_endpoint = export_id_endpoint
        self.campaign_send_endpoint = campaign_send_endpoint
        self.enterprise_campaign_id = enterprise_campaign_id
        self.from_email = from_email
        self.session = requests.Session()

    def __create_post_request(self, body, endpoint):
        """
        Creates a request and returns a response.

        Arguments:
            body (dict): The request body
            endpoint (str): The endpoint for the API e.g. /messages/send or /email/hard_bounces

        Returns:
            response (dict): The response object
        """
        message_body = json.dumps(body)
        response = {'errors': []}
        r = self._post_request(message_body, endpoint)  # pylint: disable=invalid-name
        response.update(r.json())
        response['status_code'] = r.status_code

        message = response["message"]
        response['success'] = (
            message in ('success', 'queued') and not response['errors']
        )
        if message not in ('success', 'queued'):
            raise BrazeClientError(message, response['errors'])
        return response

    def _post_request(self, body, endpoint):
        """
        Http posts the message body with associated headers.

        Arguments:
            body (dict): The request body
            endpoint (str): The endpoint for the API e.g. /messages/send or /email/hard_bounces

        Returns:
            r (requests.Response): The http response object
        """
        self.session.headers.update(
            {"Authorization": f"Bearer {self.rest_api_key}", "Content-Type": "application/json"}
        )
        r = self.session.post(urljoin(self.rest_api_url, endpoint), data=body, timeout=2)  # pylint: disable=invalid-name
        if r.status_code == 429:
            reset_epoch_s = float(r.headers.get("X-RateLimit-Reset", 0))
            raise BrazeRateLimitError(reset_epoch_s)
        if str(r.status_code).startswith('5'):
            raise BrazeInternalServerError
        return r

    def __create_get_request(self, parameters, endpoint):
        """
        Creates a request and returns a response.

        Arguments:
            parameters (dict): The request parameters
            endpoint (str): The endpoint for the API e.g. /messages/send or /email/hard_bounces

        Returns:
            response (dict): The response object
        """
        response = {'errors': []}
        r = self._get_request(parameters, endpoint)  # pylint: disable=invalid-name
        response.update(r.json())
        response['status_code'] = r.status_code
        message = response["message"]
        response['success'] = (
            message in ('success', 'queued') and not response['errors']
        )
        if message not in ('success', 'queued'):
            raise BrazeClientError(message, response['errors'])
        return response

    def _get_request(self, parameters, endpoint):
        """
        Http GET the parameters with associated headers.

        Arguments:
            parameters (dict): The request parameters
            endpoint (str): The endpoint for the API e.g. /messages/send or /email/hard_bounces

        Returns:
            r (requests.Response): The http response object
        """
        self.session.headers.update(
            {"Authorization": f"Bearer {self.rest_api_key}", "Content-Type": "application/json"}
        )
        url_with_parameters = urljoin(self.rest_api_url, endpoint) + '?' + urlencode(parameters)
        r = self.session.get(url_with_parameters)  # pylint: disable=invalid-name
        if r.status_code == 429:
            reset_epoch_s = float(r.headers.get("X-RateLimit-Reset", 0))
            raise BrazeRateLimitError(reset_epoch_s)
        if str(r.status_code).startswith('5'):
            raise BrazeInternalServerError
        return r

    def create_braze_alias(self, recipient_emails):
        """
        Creates a Braze anonymous user and assigns it the recipient email address.

        Alias naming format:
        {
            "attributes": [
                {
                    "user_alias" : {
                        "alias_name" : "Enterprise",
                        "alias_label" : "someuser@someorg.org"
                    },
                    "email" : "someuser@someorg.org"
                }
            ]
        }

        Arguments:
            recipient_emails (list): e.g. ['test1@example.com', 'test2@example.com']
        """
        if not recipient_emails:
            raise BrazeClientError('Missing parameters for Alias creation')
        user_aliases = []
        attributes = []
        for recipient_email in recipient_emails:
            if not self.get_braze_external_id(recipient_email):
                user_alias = {
                    'alias_name': 'Enterprise',
                    'alias_label': recipient_email
                }
                user_aliases.append(user_alias)
                attribute = {
                    'user_alias': {
                        'alias_name': 'Enterprise',
                        'alias_label': recipient_email
                    },
                    'email': recipient_email,
                    'is_enterprise_learner': 'true'
                }
                attributes.append(attribute)

        alias_message = {
            'user_aliases': user_aliases,
        }
        if user_aliases:
            self.__create_post_request(alias_message, self.new_alias_endpoint)

        attribute_message = {
            'attributes': attributes
        }
        if attributes:
            self.__create_post_request(attribute_message, self.users_track_endpoint)

    def send_message(  # pylint: disable=dangerous-default-value
        self,
        email_ids,
        subject,
        body,
        sender_alias='EdX Support Team',
        reply_to='',
        attachments=[],
        campaign_id='',
        message_variation_id=None,
    ):
        """
        Sends the message via Braze Rest API /messages/send
        The "override_frequency_capping" key in the request payload is important;
        it tells Braze to ignore the global campaign message frequency cap
        for the message we're sending in this method.  Since this is a transactional
        message, we'd like the recipient to receive it regardless of what/how-many
        other campaign messages they have received.

        Arguments:
            email_ids (list): e.g. ['test1@example.com', 'test2@example.com']
            subject (str): e.g. 'Test Subject'
            body (str): e.g. '<html>Test Html Message</html>'
            sender_alias (str): sender alias for email e.g. edX Support Team
            reply_to (str): Enterprise Customer reply to address for email reply
            attachments (list): list of dicts with filename and url keys
            campaign_id (str): The id of the campaign this email is associated with
            message_variation_id (str): The id of the message variant associated with the given campaign_id

        Request message format:
            Content-Type: application/json
            Authorization: Bearer {{api_key}}
            Body:
                {
                    "external_user_ids": [ "user1@example.com", "user2@example.org" ],
                    "campaign_id": "some-campaign-identifier",
                    "override_frequency_capping": true,
                    "messages": {
                        "email": {
                            "app_id": "99999999-9999-9999-9999-999999999999",
                            "subject": "Testing",
                            "from": "edX <edx-for-business-no-reply@info.edx.org>",
                            "reply_to": "reply@edx.org",
                            "body": "<html>Test</html>",
                            "message_variation_id": "email-1234",
                            "attachments": [
                               { "file_name": "filename1", "url": "url1" },
                               { "file_name": "filename2", "url": "url2" }
                            ]
                        }
                    }
                }

        Returns:
            response (dict): e.g:
                {
                    "dispatch_id": "9179894df02dab49c1d1f85c6a01c41b",
                    "message": "success"
                }
        """
        # To avoid circular import
        from ecommerce_worker.email.v1.utils import remove_special_characters_from_string  # pylint: disable=import-outside-toplevel
        if not email_ids or not subject or not body:
            raise BrazeClientError('Missing parameters for Braze email')
        self.create_braze_alias(email_ids)
        user_aliases = []
        external_ids = []
        sender_alias = remove_special_characters_from_string(sender_alias)
        for email_id in email_ids:
            external_id = self.get_braze_external_id(email_id)
            if external_id:
                external_ids.append(str(external_id))
            else:
                user_alias = {
                    'alias_name': 'Enterprise',
                    'alias_label': email_id
                }
                user_aliases.append(user_alias)
        email = {
            'app_id': self.webapp_api_key,
            'subject': subject,
            'from': sender_alias + self.from_email,
            'body': body,
        }
        if attachments:
            email['attachments'] = attachments
        if reply_to:
            email['reply_to'] = reply_to
        if message_variation_id:
            email['message_variation_id'] = message_variation_id
        message = {
            'user_aliases': user_aliases,
            'external_user_ids': external_ids,
            'recipient_subscription_state': 'all',
            'messages': {
                'email': email
            }
        }
        if campaign_id:
            message['campaign_id'] = campaign_id
            message['override_frequency_capping'] = True

        # Scrub the app_id from the log message
        cleaned_message = copy.deepcopy(message)
        cleaned_app_id = '{}...{}'.format(cleaned_message['messages']['email']['app_id'][0:4],
                                          cleaned_message['messages']['email']['app_id'][-4:])
        cleaned_message['messages']['email']['app_id'] = cleaned_app_id
        log.info(
            '[ECOMM-WORKER-BRAZE] Message: [%s], URL: [%s]', str(cleaned_message), str(self.messages_send_endpoint)
        )
        return self.__create_post_request(message, self.messages_send_endpoint)

    def did_email_bounce(
        self,
        email_id
    ):
        """
        Sends the message via Braze Rest API /email/hard_bounces

        Arguments:
            email_id (str): e.g. 'test1@example.com'

        Returns:
             True (boolean): True if email bounced,
        """
        if not email_id:
            raise BrazeClientError('Missing parameters for Braze email')
        parameters = {
            'email': email_id
        }

        response = self.__create_get_request(parameters, self.email_bounce_endpoint)
        if response["emails"]:
            return True

        return False

    def send_campaign_message(
        self,
        email_ids,
        subject,
        body,
        campaign_id='',
        sender_alias='EdX Support Team'
    ):
        """
        Sends the message via Braze Rest API /campaigns/trigger/send

        Arguments:
            campaign_id (str): The id of the API triggered campaign
            email_ids (list): e.g. ['test1@example.com', 'test2@example.com']
            subject (str): e.g. 'Test Subject'
            body (str): e.g. '<html>Test Html Message</html>'
            sender_alias (str): sender alias for email e.g. edX Support Team

        Request message format:
            Content-Type: application/json
            Authorization: Bearer {{api_key}}
            Body:
                {
                    'campaign_id': 'e2e765a1-4a2b-69b4-05ba-155a26ee2ed3',
                    'recipients': [
                        {
                            'external_user_id': 'Enterprise-someemail@someemail.com',
                            'trigger_properties': {
                                'sender_alias': "message sender alias',
                                'subject': 'message subject',
                                'body': 'message body'
                            },
                            'send_to_existing_only': false,
                            'attributes':
                                {
                                    'email' : 'someemail@someemail.com'
                                }
                        }
                    ]
                }
        Returns:
            response (dict): e.g:
                "email_1": {
                    "dispatch_id": "9179894df02dab49c1d1f85c6a01c41b",
                    "message": "success"
                },
                "email_2": {
                    "dispatch_id": "2229894df02dab49c1d1f85c6a01c432",
                    "message": "success"
                }
        """
        if not email_ids or not subject or not body:
            raise BrazeClientError('Missing parameters for Braze email')
        result = {}
        for email_id in email_ids:
            send_to_existing_only = bool(self.get_braze_external_id(email_id))
            message = {
                'campaign_id': campaign_id,
                'recipients': [
                    {
                        'external_user_id': f'Enterprise-{email_id}',
                        'trigger_properties': {
                            'sender_alias': sender_alias,
                            'subject': subject,
                            'body': body
                        },
                        'send_to_existing_only': send_to_existing_only,
                        'attributes':
                            {
                                'email': email_id
                        }
                    }
                ]
            }
            response = self.__create_post_request(message, self.campaign_send_endpoint)
            result[email_id] = response
        return result

    def get_braze_external_id(
        self,
        email_id
    ):
        """
        Checks via /users/export/ids if the user account exists in Braze

        Arguments:
            email_id (str): e.g. 'test1@example.com'

        Returns:
             external_id (int): external_id if account exists,
        """
        if not email_id:
            raise BrazeClientError('Missing parameters for Braze account check.')
        message = {
            'email_address': email_id,
            'fields_to_export': ['external_id']
        }
        response = self.__create_post_request(message, self.export_id_endpoint)
        if response["users"] and 'external_id' in response["users"][0]:
            return response["users"][0]["external_id"]

        return None


class EdxBrazeClient(edx_braze_client.BrazeClient):
    """
    Wrapper around the edx-braze-client library BrazeClient class.
    TODO: Deprecate ``BrazeClient`` above and use only this class
    for Braze interactions.
    """
    def __init__(self, site_code):
        config = get_braze_configuration(site_code)
        validate_braze_config(config, site_code)

        super().__init__(
            api_key=config.get('BRAZE_REST_API_KEY'),
            api_url=config.get('REST_API_URL'),
            app_id=config.get('BRAZE_WEBAPP_API_KEY'),
        )

    def create_recipient(
        self,
        user_email,
        lms_user_id,
        trigger_properties=None,
    ):
        """
        Create a recipient object using the given user_email and lms_user_id.
        """

        user_alias = {
            'alias_label': 'Enterprise',
            'alias_name': user_email,
        }

        # Identify the user alias in case it already exists. This is necessary so
        # we don't accidently create a duplicate Braze profile.
        self.identify_users([{
            'external_id': lms_user_id,
            'user_alias': user_alias,
        }])

        attributes = {
            "user_alias": user_alias,
            "email": user_email,
            "_update_existing_only": False,
        }

        return {
            'external_user_id': lms_user_id,
            'attributes': attributes,
            # If a profile does not already exist, Braze will create a new profile before sending a message.
            'send_to_existing_only': False,
            'trigger_properties': trigger_properties or {},
        }



"""Order fulfillment tasks."""
from ssl import SSLError
from urllib.parse import urljoin

import requests
from celery import shared_task
from celery.exceptions import Ignore
from celery.utils.log import get_task_logger
from requests.exceptions import HTTPError, Timeout

from ecommerce_worker.utils import get_configuration, get_access_token

logger = get_task_logger(__name__)


def _retry_order(self, exception, max_fulfillment_retries, order_number):
    """
    Retry with exponential backoff until fulfillment
    succeeds or the retry limit is reached. If the retry limit is exceeded,
    the exception is re-raised.
    """
    retries = self.request.retries
    if retries == max_fulfillment_retries:
        logger.exception('Fulfillment of order [%s] failed. Giving up.', order_number)
    else:
        logger.warning('Fulfillment of order [%s] failed. Retrying.', order_number)

    countdown = 2 ** retries
    raise self.retry(exc=exception, countdown=countdown, max_retries=max_fulfillment_retries)


@shared_task(bind=True, ignore_result=True)
def fulfill_order(self, order_number, site_code=None, email_opt_in=False):
    """
    Fulfills an order.

    Arguments:
        order_number (str): Order number indicating which order to fulfill.

    Returns:
        None
    """

    api_url = urljoin(
        get_configuration('ECOMMERCE_API_ROOT', site_code=site_code) + '/',
        f'orders/{order_number}/fulfill/'
    )
    max_fulfillment_retries = get_configuration('MAX_FULFILLMENT_RETRIES', site_code=site_code)
    try:
        logger.info('Requesting fulfillment of order [%s].', order_number)
        headers = {'Authorization': f'JWT {get_access_token()}'}
        params = {'email_opt_in': email_opt_in}
        response = requests.put(
            api_url,
            params=params,
            headers=headers
        )
        response.raise_for_status()

    except HTTPError as exc:
        status_code = exc.response.status_code
        if status_code == 406:
            # The order is not fulfillable. Therefore, it must be complete.
            logger.info('Order [%s] has already been fulfilled. Ignoring.', order_number)
            raise Ignore() from exc

        # Unknown connection error while fulfills an order. Let's retry to resolve it.
        logger.warning(
            'Fulfillment of order [%s] failed because of HTTPError. Retrying',
            order_number,
            exc_info=True
        )
        _retry_order(self, exc, max_fulfillment_retries, order_number)
    except (Timeout, SSLError) as exc:
        # Fulfillment failed, retry
        _retry_order(self, exc, max_fulfillment_retries, order_number)



"""
This file contains celery tasks for email marketing signal handler.
"""
import logging

from celery import shared_task

from ecommerce_worker.email.v1.braze.client import is_braze_enabled
from ecommerce_worker.email.v1.braze.tasks import (
    send_api_triggered_offer_usage_email_via_braze,
    send_code_assignment_nudge_email_via_braze,
    send_offer_assignment_email_via_braze,
    send_offer_update_email_via_braze,
    send_offer_usage_email_via_braze,
)

# Disable unused imports because these are public API functions and previous implementations used some of these
# arguments, but current ones don't use all of them. And maybe future implementations will use them again. But
# regardless, they are intentionally there as part of the public API, to be used or not by implementations.
# pylint: disable=unused-argument

logger = logging.getLogger(__name__)


@shared_task(bind=True, ignore_result=True)
def send_offer_assignment_email(self, user_email, offer_assignment_id, subject, email_body, sender_alias,  # pylint: disable=dangerous-default-value
                                reply_to='', attachments=[], site_code=None, base_enterprise_url='') -> None:
    """
    Sends the offer assignment email.

    Args:
        self: Ignore.
        user_email (str): Recipient's email address.
        offer_assignment_id (str): Key of the entry in the offer_assignment model.
        subject (str): Email subject.
        email_body (str): The body of the email.
        sender_alias (str): Enterprise Customer sender alias used as From Name.
        reply_to (str): Enterprise Customer reply to address for email reply.
        attachments (list): File attachment list with dicts having 'file_name' and 'url' keys.
        site_code (str): Identifier of the site sending the email.
        base_enterprise_url (str): Url for the enterprise learner portal.
    """
    if not is_braze_enabled(site_code):
        logger.error('Braze not enabled for site code {}'.format(site_code))
        return

    send_offer_assignment_email_via_braze(
        self, user_email, offer_assignment_id, subject, email_body,
        sender_alias, reply_to, attachments, site_code
    )


@shared_task(bind=True, ignore_result=True)
def send_offer_update_email(self, user_email, subject, email_body, sender_alias, reply_to='', attachments=[],  # pylint: disable=dangerous-default-value
                            site_code=None, base_enterprise_url='') -> None:
    """
    Sends the offer emails after assignment, either for revoking or reminding.

    Args:
        self: Ignore.
        user_email (str): Recipient's email address.
        subject (str): Email subject.
        email_body (str): The body of the email.
        site_code (str): Identifier of the site sending the email.
        sender_alias (str): Enterprise Customer sender alias used as From Name.
        reply_to (str): Enterprise Customer reply to address for email reply.
        attachments (list): File attachment list with dicts having 'file_name' and 'url' keys.
        base_enterprise_url (str): Enterprise learner portal url.
    """
    if not is_braze_enabled(site_code):
        logger.error('Braze not enabled for site code {}'.format(site_code))
        return

    send_offer_update_email_via_braze(
        self, user_email, subject, email_body, sender_alias, reply_to, attachments, site_code
    )


@shared_task(bind=True, ignore_result=True)
def send_offer_usage_email(
    self, emails, subject, email_body, reply_to='', attachments=None, site_code=None, base_enterprise_url=''
) -> None:
    """
    DEPRECATED: Will soon be replaced with ``send_api_triggered_offer_usage_email`` below.
    See https://2u-internal.atlassian.net/browse/ENT-5940

    Sends the offer usage email.
    Args:
        self: Ignore.
        emails (str): comma separated emails.
        subject (str): Email subject.
        email_body (str): The body of the email.
        reply_to (str): Enterprise Customer reply to address for email reply.
        attachments (list): File attachment list with dicts having 'file_name' and 'url' keys.
        site_code (str): Identifier of the site sending the email.
        base_enterprise_url (str): Url of the enterprise's learner portal
    """
    if is_braze_enabled(site_code):
        send_offer_usage_email_via_braze(
            self, emails, subject, email_body, reply_to, attachments or [], site_code,
        )


@shared_task(bind=True, ignore_result=True)
def send_api_triggered_offer_usage_email(
    self, lms_user_ids_by_email, subject, email_body_variables, site_code=None, campaign_id=None
) -> None:
    """
    Sends a "digest" email to recipients defined on a ConditionalOffer
    describing how much of the offer has been used.

    Args:
        self: Ignore.
        lms_user_ids_by_email (dict): Map of user email to LMS user ids.
        subject (str): Email subject.
        email_body_variables (dict): key-value pairs that are injected into Braze email template for personalization.
        site_code (str): Identifier of the site sending the email.
        campaign_id (str): Identifier of Braze API-triggered campaign to send message through; defaults
            to config.ENTERPRISE_CODE_USAGE_CAMPAIGN_ID
    """
    if not is_braze_enabled(site_code):
        logger.error('Braze not enabled for site code {}'.format(site_code))
        return

    send_api_triggered_offer_usage_email_via_braze(
        self, lms_user_ids_by_email, subject, email_body_variables, site_code, campaign_id,
    )


@shared_task(bind=True, ignore_result=True)
def send_code_assignment_nudge_email(self, email, subject, email_body, sender_alias, reply_to='', attachments=[],  # pylint: disable=dangerous-default-value
                                     site_code=None, base_enterprise_url='') -> None:
    """
    Sends the code assignment nudge email.

    Args:
        self: Ignore.
        email (str): Recipient's email address.
        subject (str): Email subject.
        email_body (str): The body of the email.
        sender_alias (str): Enterprise Customer sender alias used as From Name.
        reply_to (str): Enterprise Customer reply to address for email reply.
        attachments (list): File attachment list with dicts having 'file_name' and 'url' keys.
        site_code (str): Identifier of the site sending the email.
        base_enterprise_url (str): Enterprise learner portal url.
    """
    if not is_braze_enabled(site_code):
        logger.error('Braze not enabled for site code {}'.format(site_code))
        return

    send_code_assignment_nudge_email_via_braze(
        self, email, subject, email_body, sender_alias, reply_to, attachments, site_code
    )



""" Utility functions. """
import string
from json import JSONDecodeError
from urllib.parse import urljoin

import requests
from requests.exceptions import HTTPError

from ecommerce_worker.email.v1.braze.client import is_braze_enabled, get_braze_client
from ecommerce_worker.utils import get_access_token, get_configuration


def update_assignment_email_status(offer_assignment_id, send_id, status, site_code=None):
    """
    Update the offer_assignment and offer_assignment_email model using the Ecommerce assignment-email api.
    Arguments:
        offer_assignment_id (str): Key of the entry in the offer_assignment model.
        send_id (str): Unique message id from Sailthru
        status (str): status to be sent to the api
        site_code (str): site code
    Returns:
        True or False based on model update status from Ecommerce api
    """

    api_url = urljoin(get_configuration('ECOMMERCE_API_ROOT', site_code=site_code) + '/', 'assignment-email/status/')
    try:
        headers = {'Authorization': f'JWT {get_access_token()}'}
        post_data = {
            'offer_assignment_id': offer_assignment_id,
            'send_id': send_id,
            'status': status,
        }
        response = requests.post(
            api_url,
            data=post_data,
            headers=headers
        )
        response.raise_for_status()
        data = response.json()
    except (HTTPError, JSONDecodeError):
        return False

    return bool(data.get('status') == 'updated')


def did_email_bounce(user_email, site_code=None) -> bool:
    """
    Checks if the given user's emails have bounced.

    Args:
        user_email (str): Recipient's email address.
        site_code (str): Identifier of the site sending the email.
    """
    if is_braze_enabled():  # pylint: disable=no-value-for-parameter
        client = get_braze_client(site_code)
        return client.did_email_bounce(user_email)

    return False


def remove_special_characters_from_string(input_string):
    """
    Removes all special characters from the string passed as argument.
    string.punctuation contains the following characters:  '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'

    Args:
        input_string (str): Input string from which special characters need to be removed.
    """
    if input_string:
        return input_string.translate(str.maketrans('', '', string.punctuation))
    return ''



"""
Tests for Braze client.
"""

from unittest import mock, TestCase
from unittest.mock import patch, Mock
from urllib.parse import urlencode

import ddt
import responses

from ecommerce_worker.email.v1.braze.client import get_braze_client, EdxBrazeClient
from ecommerce_worker.email.v1.braze.exceptions import (
    BrazeClientError,
    BrazeNotEnabled,
    BrazeRateLimitError,
    BrazeInternalServerError,
    ConfigurationError
)


SITE_CODE = 'test'
BRAZE_OVERRIDES = {
    SITE_CODE: {
        'BRAZE': {
            'BRAZE_ENABLE': True,
            'BRAZE_REST_API_KEY': 'rest_api_key',
            'BRAZE_WEBAPP_API_KEY': 'webapp_api_key',
            'REST_API_URL': 'https://rest.iad-06.braze.com',
            'MESSAGES_SEND_ENDPOINT': '/messages/send',
            'EMAIL_BOUNCE_ENDPOINT': '/email/hard_bounces',
            'NEW_ALIAS_ENDPOINT': '/users/alias/new',
            'USERS_TRACK_ENDPOINT': '/users/track',
            'EXPORT_ID_ENDPOINT': '/users/export/ids',
            'CAMPAIGN_SEND_ENDPOINT': '/campaigns/trigger/send',
            'ENTERPRISE_CAMPAIGN_ID': '',
            'FROM_EMAIL': '<edx-for-business-no-reply@info.edx.org>',
        }
    }
}


@ddt.ddt
class BrazeClientTests(TestCase):
    """
    Tests for Braze Client.
    """
    def mock_braze_user_endpoints(self, users=[]):  # pylint: disable=dangerous-default-value
        """ Mock POST requests to the user alias, track and export endpoints. """
        host = 'https://rest.iad-06.braze.com/users/track'
        responses.add(
            responses.POST,
            host,
            json={'message': 'success'},
            status=201
        )
        host = 'https://rest.iad-06.braze.com/users/alias/new'
        responses.add(
            responses.POST,
            host,
            json={'message': 'success'},
            status=201
        )
        host = 'https://rest.iad-06.braze.com/users/export/ids'
        responses.add(
            responses.POST,
            host,
            json={"users": users, "message": "success"},
            status=201
        )

    def assert_get_braze_client_raises(self, exc_class, config):
        """
        Asserts an error is raised by a call to get_braze_client.
        """
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=config)):
            with self.assertRaises(exc_class):
                get_braze_client(SITE_CODE)

    def test_get_braze_client_with_braze_disabled(self):
        """
        Verify the method raises a BrazeNotEnabled if Braze is not enabled for the site.
        """

        with mock.patch('ecommerce_worker.email.v1.braze.client.log.error') as mock_log_error:
            self.assert_get_braze_client_raises(BrazeNotEnabled, {'BRAZE_ENABLE': False})

        mock_log_error.assert_called_once_with(f'Braze is not enabled for site {SITE_CODE}')

    @ddt.data(
        {},
        {'BRAZE_REST_API_KEY': None, 'BRAZE_WEBAPP_API_KEY': None},
        {'BRAZE_REST_API_KEY': 'test', 'BRAZE_WEBAPP_API_KEY': None},
    )
    def test_get_braze_client_without_credentials(self, braze_config):
        """
        Verify the method raises a ConfigurationError if Braze is not configured properly.
        """
        braze_config['BRAZE_ENABLE'] = True

        with mock.patch('ecommerce_worker.email.v1.braze.client.log.error') as mock_log:
            self.assert_get_braze_client_raises(ConfigurationError, braze_config)

        mock_log.assert_called_once_with(f'Required keys missing for site {SITE_CODE}')

    def test_create_braze_alias(self):
        """
        Asserts an error is raised by a call to create_braze_alias.
        """
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            with self.assertRaises(BrazeClientError):
                client.create_braze_alias(recipient_emails=[])

    @responses.activate
    def test_send_braze_message_success(self):
        """
        Verify that an email message is sent via /messages/send.
        """
        self.mock_braze_user_endpoints()
        success_response = {
            'dispatch_id': '66cdc28f8f082bc3074c0c79f',
            'errors': [],
            'message': 'success',
            'status_code': 201
        }
        host = 'https://rest.iad-06.braze.com/messages/send'
        responses.add(
            responses.POST,
            host,
            json=success_response,
            status=201
        )
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            response = client.send_message(
                email_ids=['test1@example.com', 'test2@example.com'],
                subject='Test Subject',
                body='<html>Test Html Message</html>',
                attachments=[{'file_name': 'filename1', 'url': 'url1'}]
            )
            self.assertEqual(response['success'], True)

    @responses.activate
    def test_send_braze_message_success_with_external_ids(self):
        """
        Verify that an email message is sent via /messages/send in presence of existing edX users.
        """
        self.mock_braze_user_endpoints(users=[{"external_id": "99999"}])
        success_response = {
            'dispatch_id': '66cdc28f8f082bc3074c0c79f',
            'errors': [],
            'message': 'success',
            'status_code': 201
        }
        host = 'https://rest.iad-06.braze.com/messages/send'
        responses.add(
            responses.POST,
            host,
            json=success_response,
            status=201
        )
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            response = client.send_message(
                ['test1@example.com', 'test2@example.com'],
                'Test Subject',
                '<html>Test Html Message</html>'
            )
            self.assertEqual(response['success'], True)

    @responses.activate
    @ddt.data(
        (400, BrazeClientError),
        (429, BrazeRateLimitError),
        (500, BrazeInternalServerError),
    )
    @ddt.unpack
    def test_send_braze_message_failure(self, status_code, error):
        """
        Verify that a failed email message throws the relevant error.
        """
        self.mock_braze_user_endpoints()
        failure_response = {
            'message': 'Not a Success',
            'status_code': status_code
        }
        host = 'https://rest.iad-06.braze.com/messages/send'

        responses.add(
            responses.POST,
            host,
            json=failure_response,
            status=status_code
        )
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            with self.assertRaises(error):
                response = client.send_message(  # pylint: disable=unused-variable
                    ['test1@example.com', 'test2@example.com'],
                    'Test Subject',
                    '<html>Test Html Message</html>'
                )

    @ddt.data(
        (['test1@example.com', 'test2@example.com'], None, '<html>Test Html Message</html>'),
        (None, 'Test Subject', '<html>Test Html Message</html>'),
        (['test1@example.com', 'test2@example.com'], 'Test Subject', None),
    )
    @ddt.unpack
    def test_send_braze_message_failure_missing_parameters(self, email, subject, body):
        """
        Verify that an error is raised for missing email parameters.
        """
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            with self.assertRaises(BrazeClientError):
                response = client.send_message(email, subject, body)  # pylint: disable=unused-variable

    @responses.activate
    def test_send_braze_campaign_message_success(self):
        """
        Verify that an email message is sent via /campaigns/trigger/send.
        """
        self.mock_braze_user_endpoints()
        success_response = {
            'dispatch_id': '66cdc28f8f082bc3074c0c79f',
            'errors': [],
            'message': 'success',
            'status_code': 201
        }
        host = 'https://rest.iad-06.braze.com/campaigns/trigger/send'
        responses.add(
            responses.POST,
            host,
            json=success_response,
            status=201
        )
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            response = client.send_campaign_message(
                ['test1@example.com', 'test2@example.com'],
                'Test Subject',
                '<html>Test Html Message</html>'
            )
            self.assertEqual(len(response.keys()), 2)

    @responses.activate
    @ddt.data(
        (400, BrazeClientError),
        (429, BrazeRateLimitError),
        (500, BrazeInternalServerError),
    )
    @ddt.unpack
    def test_send_braze_campaign_message_failure(self, status_code, error):
        """
        Verify that a failed email message throws the relevant error.
        """
        self.mock_braze_user_endpoints()
        failure_response = {
            'message': 'Not a Success',
            'status_code': status_code
        }
        host = 'https://rest.iad-06.braze.com/campaigns/trigger/send'

        responses.add(
            responses.POST,
            host,
            json=failure_response,
            status=status_code
        )
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            with self.assertRaises(error):
                response = client.send_campaign_message(  # pylint: disable=unused-variable
                    ['test1@example.com', 'test2@example.com'],
                    'Test Subject',
                    '<html>Test Html Message</html>'
                )

    @ddt.data(
        (['test1@example.com', 'test2@example.com'], None, '<html>Test Html Message</html>'),
        (None, 'Test Subject', '<html>Test Html Message</html>'),
        (['test1@example.com', 'test2@example.com'], 'Test Subject', None),
    )
    @ddt.unpack
    def test_send_braze_campaign_message_failure_missing_parameters(self, email, subject, body):
        """
        Verify that an error is raised for missing email parameters.
        """
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            with self.assertRaises(BrazeClientError):
                response = client.send_campaign_message(email, subject, body)  # pylint: disable=unused-variable

    @ddt.data(
        (
            {
                'emails': [
                    {
                        'email': 'foo@braze.com',
                        'hard_bounced_at': '2016-08-25 15:24:32 +0000'
                    }
                ],
                'message': 'success',
            },
            True
        ),
        (
            {
                'emails': [],
                'message': 'success',
            },
            False
        ),
    )
    @ddt.unpack
    @responses.activate
    def test_bounced_email(self, response, did_bounce):
        """
        Verify that an email id bounced on Braze.
        """
        bounced_email = 'foo@braze.com'
        host = 'https://rest.iad-06.braze.com/email/hard_bounces?{email}'.format(
            email=urlencode({'email': bounced_email})
        )

        responses.add(
            responses.GET,
            host,
            json=response,
            status=200
        )
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            bounce = client.did_email_bounce(bounced_email)
            self.assertEqual(bounce, did_bounce)

    @responses.activate
    @ddt.data(
        (400, BrazeClientError),
        (429, BrazeRateLimitError),
        (500, BrazeInternalServerError),
    )
    @ddt.unpack
    def test_bounce_message_failure(self, status_code, error):
        """
        Verify that a failed email message throws the relevant error.
        """
        failure_response = {
            'message': 'Not a Success'
        }
        bounced_email = 'foo@braze.com'
        host = 'https://rest.iad-06.braze.com/email/hard_bounces?{email}'.format(
            email=urlencode({'email': bounced_email})
        )

        responses.add(
            responses.GET,
            host,
            json=failure_response,
            status=status_code
        )
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            with self.assertRaises(error):
                client.did_email_bounce(bounced_email)

    def test_did_email_bounce_failure_missing_parameters(self):
        """
        Verify that an error is raised for missing parameters.
        """
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            with self.assertRaises(BrazeClientError):
                client.did_email_bounce(email_id=None)

    def test_get_braze_external_id(self):
        """
        Asserts an error is raised by a call to get_braze_external_id.
        """
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            with self.assertRaises(BrazeClientError):
                client.get_braze_external_id(email_id=None)

    @responses.activate
    def test_get_braze_external_id_success(self):
        """
        Tests a successful call to the /users/export/ids endpoint
        """
        host = 'https://rest.iad-06.braze.com/users/export/ids'
        responses.add(
            responses.POST,
            host,
            json={"users": [{"external_id": "5261613"}], "message": "success"},
            status=201
        )
        braze = BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
        with patch('ecommerce_worker.email.v1.braze.client.get_braze_configuration', Mock(return_value=braze)):
            client = get_braze_client(SITE_CODE)
            external_id = client.get_braze_external_id(email_id='test@example.com')
            self.assertEqual(external_id, '5261613')


@ddt.ddt
class EdxBrazeClientTests(TestCase):
    """
    Tests for the EdxBrazeClient wrapper class.
    """
    @ddt.data(
        (
            {'BRAZE_ENABLE': False},
            BrazeNotEnabled,
        ),
        (
            {
                'BRAZE_REST_API_KEY': 'some-key',
                'BRAZE_ENABLE': True,
            },
            ConfigurationError,
        ),
    )
    @ddt.unpack
    def test_braze_config_invalid(self, mock_config, expected_exception):
        """
        Asserts an error is raised when initializing an EdxBrazeClient for a config
        that has Braze disabled or is missing a required key.
        """
        with patch(
            'ecommerce_worker.email.v1.braze.client.get_braze_configuration',
            Mock(return_value=mock_config),
        ) as mock_braze_config:
            with self.assertRaises(expected_exception):
                EdxBrazeClient(SITE_CODE)

            mock_braze_config.assert_called_once_with(SITE_CODE)

    @patch(
        'ecommerce_worker.email.v1.braze.client.get_braze_configuration',
        return_value=BRAZE_OVERRIDES[SITE_CODE]['BRAZE']
    )
    def test_create_recipient(self, mock_get_braze_config):  # pylint: disable=unused-argument
        """
        Tests that create_recipient() calls the underlying identify_users() method
        and returns a dictionary with expected keys and values.
        """
        user_email = 'curly@stooges.org'
        lms_user_id = 777
        trigger_properties = {'foo': 12}
        with patch(
            'ecommerce_worker.email.v1.braze.client.EdxBrazeClient.identify_users'
        ) as mock_identify_users:
            client = EdxBrazeClient(SITE_CODE)
            actual_recipient = client.create_recipient(user_email, lms_user_id, trigger_properties)

            expected_recipient = {
                'attributes': {
                    '_update_existing_only': False,
                    'email': 'curly@stooges.org',
                    'user_alias': {
                        'alias_label': 'Enterprise',
                        'alias_name': 'curly@stooges.org'
                    }
                },
                'external_user_id': 777,
                'send_to_existing_only': False,
                'trigger_properties': {
                    'foo': 12,
                },
            }
            self.assertEqual(actual_recipient, expected_recipient)



"""
Braze-related exception classes.
"""


class BrazeError(Exception):
    """
    Base class for Braze-related errors.
    """


class ConfigurationError(BrazeError):
    """
    Raised when Braze is not properly configured.
    """


class BrazeNotEnabled(BrazeError):
    """
    Raised when Braze is not enabled.
    """


class BrazeClientError(BrazeError):
    """
    Represents any Braze Client Error.
    https://www.braze.com/docs/developer_guide/rest_api/user_data/#user-track-responses
    """


class BrazeRateLimitError(BrazeClientError):
    """A rate limit error was encountered."""
    def __init__(self, reset_epoch_s):
        """
        Arguments:
            reset_epoch_s (float): Unix timestamp for when the API may be called again.
        """
        self.reset_epoch_s = reset_epoch_s
        super().__init__()


class BrazeInternalServerError(BrazeClientError):
    """
    Used for Braze API responses where response code is of type 5XX suggesting
    Braze side server errors.
    """



""" Tests for utility functions. """
import json
from json import JSONDecodeError
from unittest import TestCase, mock
from urllib.parse import urljoin

import ddt
import responses
from requests.exceptions import HTTPError

from ecommerce_worker.email.v1.utils import (
    did_email_bounce,
    remove_special_characters_from_string,
    update_assignment_email_status,
)
from ecommerce_worker.utils import get_configuration


@ddt.ddt
class EmailUtilsTests(TestCase):
    """
    Tests for email v1 utility functions.
    """

    OAUTH_ACCESS_TOKEN_URL = urljoin(
        get_configuration('BACKEND_SERVICE_EDX_OAUTH2_PROVIDER_URL') + '/', 'access_token/'
    )
    ACCESS_TOKEN = 'FAKE-access-token'

    def mock_ecommerce_assignment_email_api(self, body, status=200):
        """ Mock POST requests to the ecommerce assignment-email API endpoint. """
        responses.reset()
        responses.add(
            responses.POST, '{}/assignment-email/status/'.format(
                get_configuration('ECOMMERCE_API_ROOT').strip('/')
            ),
            status=status,
            body=json.dumps(body), content_type='application/json',
        )

    def mock_access_token_api(self, status=200):
        """
        Mock POST requests to retrieve an access token for this site's service user.
        """
        responses.add(
            responses.POST,
            self.OAUTH_ACCESS_TOKEN_URL,
            status=status,
            json={
                'access_token': self.ACCESS_TOKEN
            }
        )

    @responses.activate
    @ddt.data(
        (
            # Success case
            {
                'offer_assignment_id': '555',
                'send_id': '1234ABC',
                'status': 'updated',
                'error': ''
            },
            True,
        ),
        (
            # Exception case
            {
                'offer_assignment_id': '555',
                'send_id': '1234ABC',
                'status': 'failed',
                'error': ''
            },
            False,
        ),
    )
    @ddt.unpack
    def test_update_assignment_email_status(self, data, return_value):
        """
        Test routine that updates email send status in ecommerce.
        """
        self.mock_ecommerce_assignment_email_api(data)
        self.mock_access_token_api()
        self.assertEqual(
            update_assignment_email_status('555', '1234ABC', 'success'),
            return_value
        )

    @ddt.data(HTTPError, JSONDecodeError)
    @mock.patch('ecommerce_worker.email.v1.utils.get_access_token')
    @mock.patch('ecommerce_worker.email.v1.utils.requests.post')
    def test_update_assignment_email_exceptions(self, expected_exception, mock_access_token, mock_request_post):
        """
        Test that we gracefully catch a request exceptions.
        """
        mock_access_token.return_value.json.return_value = {'access_token': 'FAKE-access-token'}

        mock_response = mock.MagicMock()
        mock_response.raise_for_status.side_effect = expected_exception
        mock_request_post.return_value = mock_response

        assert not update_assignment_email_status('555', '1234ABC', 'success')

    @mock.patch('ecommerce_worker.email.v1.utils.get_braze_client')
    @mock.patch('ecommerce_worker.email.v1.utils.is_braze_enabled')
    @ddt.data(True, False)
    def test_did_email_bounce_routing(self, braze_enabled, braze_enabled_mock, braze_client_mock):
        """
        Test that we direct requests for email bouncing correctly.
        """
        braze_enabled_mock.return_value = braze_enabled
        braze_client_mock.return_value.did_email_bounce.return_value = True
        bounced = did_email_bounce('email', 'site_code')
        assert bounced == braze_enabled

    @ddt.data(
        ('', ''),
        ('EdX Support Team', 'EdX Support Team'),
        ('ABC Mills, Inc. on edX', 'ABC Mills Inc on edX'),
        ('A&B Enterprise, inc.', 'AB Enterprise inc'),
    )
    @ddt.unpack
    def test_remove_special_characters_from_string(self, input_string, expected):
        """
        Test that the utility returns string with all special characters removed.
        """
        actual_string = remove_special_characters_from_string(input_string)
        self.assertEqual(actual_string, expected)



"""Public API for ecommerce_worker.email.v1"""

# Disable unused imports because these imports are meant for external consumption
# pylint: disable=unused-import

# Celery tasks
from .tasks import (
    send_api_triggered_offer_usage_email,
    send_code_assignment_nudge_email,
    send_offer_assignment_email,
    send_offer_update_email,
    send_offer_usage_email,
)

# Normal helper methods
from .utils import (
    did_email_bounce,
)



"""Tests of fulfillment tasks."""
# pylint: disable=no-value-for-parameter
from unittest import TestCase, mock
from urllib.parse import urljoin

import ddt
import responses
from celery.exceptions import Ignore
from edx_rest_api_client import exceptions
from requests.exceptions import HTTPError

# Ensures that a Celery app is initialized when tests are run.
from ecommerce_worker import celery_app  # pylint: disable=unused-import
from ecommerce_worker.fulfillment.v1.tasks import fulfill_order
from ecommerce_worker.utils import get_configuration


@ddt.ddt
class OrderFulfillmentTaskTests(TestCase):
    """
    Tests of the order fulfillment task.
    """
    ORDER_NUMBER = 'FAKE-123456'
    API_URL = urljoin(get_configuration('ECOMMERCE_API_ROOT') + '/', f'orders/{ORDER_NUMBER}/fulfill/')
    OAUTH_ACCESS_TOKEN_URL = urljoin(
        get_configuration('BACKEND_SERVICE_EDX_OAUTH2_PROVIDER_URL') + '/', 'access_token/'
    )
    ACCESS_TOKEN = 'FAKE-access-token'

    def setUp(self):
        super().setUp()
        self.mock_access_token_api()

    def mock_access_token_api(self, status=200):
        """
        Mock POST requests to retrieve an access token for this site's service user.
        """
        responses.add(
            responses.POST,
            self.OAUTH_ACCESS_TOKEN_URL,
            status=status,
            json={
                'access_token': self.ACCESS_TOKEN
            }
        )

    @ddt.data(
        'ECOMMERCE_API_ROOT',
        'MAX_FULFILLMENT_RETRIES',
        'BACKEND_SERVICE_EDX_OAUTH2_PROVIDER_URL',
        'BACKEND_SERVICE_EDX_OAUTH2_KEY',
        'BACKEND_SERVICE_EDX_OAUTH2_SECRET'
    )
    def test_requires_configuration(self, setting):
        """
        Verify that the task refuses to run without the configuration it requires.
        """
        with mock.patch('ecommerce_worker.configuration.test.' + setting, None):
            with self.assertRaises(RuntimeError):
                fulfill_order(self.ORDER_NUMBER)

    @responses.activate
    def test_fulfillment_success(self):
        """
        Verify that the task exits without an error when fulfillment succeeds.
        """
        responses.add(responses.PUT, self.API_URL, status=200, body="{}")

        result = fulfill_order.delay(self.ORDER_NUMBER).get()
        self.assertIsNone(result)

        # Validate the value of the HTTP Authorization header.
        last_request = responses.calls[-1].request
        token = last_request.headers.get('Authorization').split()[1]
        self.assertEqual(token, self.ACCESS_TOKEN)

    @responses.activate
    def test_fulfillment_not_possible(self):
        """
        Verify that the task exits without an error when fulfillment is not possible.
        """
        responses.add(responses.PUT, self.API_URL, status=406, body="{}")

        with self.assertRaises(Ignore):
            fulfill_order(self.ORDER_NUMBER)

    @responses.activate
    def test_fulfillment_connection_error(self):
        """
        Verify that the task raises an exception when fulfillment fails because of connection error.
        """
        responses.add(responses.PUT, self.API_URL, status=404, body="{}")

        with self.assertRaises(HTTPError):
            fulfill_order(self.ORDER_NUMBER)

    @responses.activate
    def test_fulfillment_connection_error_retry_success(self):
        """
        Verify that the task is capable of successfully retrying after connection error.
        """
        responses.add(
            responses.Response(responses.PUT, self.API_URL, status=404, body="{}"),
        )
        responses.add(
            responses.Response(responses.PUT, self.API_URL, status=200, body="{}"),
        )

        result = fulfill_order.delay(self.ORDER_NUMBER).get()
        self.assertIsNone(result)

    @responses.activate
    def test_fulfillment_failure(self):
        """
        Verify that the task raises an exception when fulfillment fails.
        """
        responses.add(responses.PUT, self.API_URL, status=500, body="{}")

        with self.assertRaises(HTTPError):
            fulfill_order.delay(self.ORDER_NUMBER).get()

    @responses.activate
    def test_fulfillment_timeout(self):
        """
        Verify that the task raises an exception when fulfillment times out.
        """
        responses.add(responses.PUT, self.API_URL, status=404, body=exceptions.Timeout())

        with self.assertRaises(exceptions.Timeout):
            fulfill_order.delay(self.ORDER_NUMBER).get()

    @responses.activate
    def test_fulfillment_retry_success(self):
        """
        Verify that the task is capable of successfully retrying after fulfillment failure.
        """
        responses.add(
            responses.Response(responses.PUT, self.API_URL, status=500, body="{}"),
        )
        responses.add(
            responses.Response(responses.PUT, self.API_URL, status=200, body="{}"),
        )

        result = fulfill_order.delay(self.ORDER_NUMBER).get()
        self.assertIsNone(result)

    @ddt.data(
        [True, 'True'],
        [False, 'False']
    )
    @ddt.unpack
    @responses.activate
    def test_email_opt_in_parameter_sent(self, email_opt_in_bool, email_opt_in_str):
        """
        Verify that the task correctly adds the email_opt_in parameter to the request.
        """
        email_opt_in_api_url = self.API_URL + '?email_opt_in=' + email_opt_in_str
        responses.add(responses.PUT, email_opt_in_api_url, status=200, body="{}")

        result = fulfill_order.delay(self.ORDER_NUMBER, email_opt_in=email_opt_in_bool).get()
        self.assertIsNone(result)

        last_request = responses.calls[-1].request
        self.assertIn('?email_opt_in=' + email_opt_in_str, last_request.path_url)



"""
This file contains a primitive cache
"""
import threading
import time

lock = threading.Lock()


class CacheObject:
    """Object saved in cache"""
    def __init__(self, value, duration):
        self.value = value
        self.expire = time.time() + duration


class Cache(dict):
    """
    Primitive key/value cache.  Entries are kept in a dict with an expiration.
    When a get of an expired entry is done, the cache is cleaned of all expired entries.
    Locking is used for thread safety
    """
    def get(self, key):
        """Get an object from the cache

        Arguments:
            key (str): Cache key

        Returns:
            Cached object
        """
        lock.acquire()  # pylint: disable=consider-using-with
        try:
            if key not in self:
                return None

            current_time = time.time()
            if self[key].expire > current_time:
                return self[key].value

            # expired key, clean out all expired keys
            deletes = []
            for k, val in self.items():
                if val.expire <= current_time:
                    deletes.append(k)
            for k in deletes:
                del self[k]

            return None
        finally:
            lock.release()

    def set(self, key, value, duration):
        """Save an object in the cache

        Arguments:
            key (str): Cache key
            value (object): object to cache
            duration (int): time in seconds to keep object in cache

        """
        lock.acquire()  # pylint: disable=consider-using-with
        try:
            self[key] = CacheObject(value, duration)
        finally:
            lock.release()



""" Tests for api entry point. """

from inspect import getmembers
from unittest import TestCase

from ecommerce_worker.email.v1 import api


class EmailApiTests(TestCase):
    """Tests for email v1 api entry point."""

    def test_public_api(self):
        """
        Test that we don't accidentally expose anything we don't expect to
        """
        assert [k for k, v in getmembers(api) if not k.startswith('_')] == [
            'did_email_bounce',
            'send_api_triggered_offer_usage_email',
            'send_code_assignment_nudge_email',
            'send_offer_assignment_email',
            'send_offer_update_email',
            'send_offer_usage_email',
        ]



import os


# Environment variable indicating which configuration module to use
# when running the worker.
CONFIGURATION_MODULE = 'WORKER_CONFIGURATION_MODULE'


def get_overrides_filename(variable):
    """
    Get the name of the file containing configuration overrides
    from the provided environment variable.
    """
    filename = os.environ.get(variable)

    if filename is None:
        msg = f'Please set the {variable} environment variable.'
        raise EnvironmentError(msg)

    return filename



from logging.config import dictConfig

from ecommerce_worker.configuration.base import *
from ecommerce_worker.configuration.logger import get_logger_config


# CELERY
CELERY_ALWAYS_EAGER = True
# END CELERY


# ORDER FULFILLMENT
ECOMMERCE_API_ROOT = 'http://localhost:18130/api/v2/'
WORKER_ACCESS_TOKEN = 'fake-access-token'
# END ORDER FULFILLMENT


# LOGGING
logger_config = get_logger_config(debug=True, dev_env=True, local_loglevel='DEBUG')
dictConfig(logger_config)
# END LOGGING


# AUTHENTICATION
ECOMMERCE_SERVICE_USERNAME = 'service'
# END AUTHENTICATION


# SITE-SPECIFIC CONFIGURATION OVERRIDES
SITE_OVERRIDES = {}



import logging
from logging.config import dictConfig

from ecommerce_worker.configuration.base import *
from ecommerce_worker.configuration.logger import get_logger_config

logger = logging.getLogger(__name__)


# CELERY
BROKER_URL = 'redis://'
# END CELERY


# ORDER FULFILLMENT
ECOMMERCE_API_ROOT = 'http://localhost:18130/api/v2/'
# END ORDER FULFILLMENT

# LOGGING
logger_config = get_logger_config(debug=True, dev_env=True, local_loglevel='DEBUG')
dictConfig(logger_config)
# END LOGGING

# Apply any developer-defined overrides.
try:
    from ecommerce_worker.configuration.private import *  # pylint: disable=import-error
except ImportError:
    logger.warning('No developer-defined configuration overrides have been applied.')
    pass



"""Helper functions."""
import os
import sys
from urllib.parse import urljoin

import requests
from edx_rest_api_client.client import REQUEST_CONNECT_TIMEOUT, REQUEST_READ_TIMEOUT
from requests.exceptions import HTTPError, RequestException

from ecommerce_worker.configuration import CONFIGURATION_MODULE


def get_configuration(variable, site_code=None):
    """
    Get a value from configuration.

    Retrieves the value corresponding to the given variable from the configuration module
    currently in use by the app.  Specify a site_code value to check for a site-specific override.

    Arguments:
        variable (str): The name of a variable from the configuration module.

    Keyword Arguments:
        site_code (str): The SITE_OVERRIDES key to inspect for site-specific values

    Returns:
        The value corresponding to the variable, or None if the variable is not found.
    """
    name = os.environ.get(CONFIGURATION_MODULE)

    # __import__ performs a full import, but only returns the top-level
    # package, not the targeted module. sys.modules is a dictionary
    # mapping module names to loaded modules.
    __import__(name)
    module = sys.modules[name]

    # Locate the setting in the specified module, then attempt to apply a site-specific override
    setting_value = getattr(module, variable, None)
    site_overrides = getattr(module, 'SITE_OVERRIDES', None)
    if site_overrides and site_code is not None:
        site_specific_overrides = site_overrides.get(site_code)
        if site_specific_overrides:
            override_value = site_specific_overrides.get(variable)
            if override_value:
                setting_value = override_value

    if setting_value is None:
        raise RuntimeError(f'Worker is improperly configured: {variable} is unset in {module}.')
    return setting_value


def get_access_token():
    """
    Returns an access token for this site's service user.

    Returns:
        str: JWT access token
    """
    oauth_access_token_url = urljoin(
        get_configuration('BACKEND_SERVICE_EDX_OAUTH2_PROVIDER_URL') + '/', 'access_token/'
    )
    try:
        post_data = {
            'grant_type': 'client_credentials',
            'client_id': get_configuration('BACKEND_SERVICE_EDX_OAUTH2_KEY'),
            'client_secret': get_configuration('BACKEND_SERVICE_EDX_OAUTH2_SECRET'),
            'token_type': 'jwt',
        }
        headers = {
            'User-Agent': 'ecommerce-worker',
        }
        response = requests.post(
            oauth_access_token_url,
            data=post_data,
            headers=headers,
            timeout=(REQUEST_CONNECT_TIMEOUT, REQUEST_READ_TIMEOUT)
        )
        response.raise_for_status()
        data = response.json()
        access_token = data['access_token']
    except (HTTPError, KeyError) as exc:
        raise RequestException(response=response) from exc

    return access_token



"""
This file contains celery task functionality for braze.
"""
from operator import itemgetter

import braze.exceptions as edx_braze_exceptions
from celery.utils.log import get_task_logger

from ecommerce_worker.email.v1.braze.client import (
    get_braze_client,
    get_braze_configuration,
    EdxBrazeClient,
)
from ecommerce_worker.email.v1.braze.exceptions import BrazeError, BrazeRateLimitError, BrazeInternalServerError
from ecommerce_worker.email.v1.utils import update_assignment_email_status

logger = get_task_logger(__name__)

# Use a smaller countdown for the offer usage task,
# since the mgmt command that executes it blocks until the task is done/failed.
OFFER_USAGE_RETRY_DELAY_SECONDS = 10


def send_offer_assignment_email_via_braze(self, user_email, offer_assignment_id, subject, email_body, sender_alias,
                                          reply_to, attachments, site_code):
    """
    Sends the offer assignment email via Braze.

    Args:
        self: Ignore.
        user_email (str): Recipient's email address.
        offer_assignment_id (str): Key of the entry in the offer_assignment model.
        subject (str): Email subject.
        email_body (str): The body of the email.
        sender_alias (str): Enterprise Customer sender alias used as From Name.
        reply_to (str): Enterprise Customer reply to address for email reply.
        attachments (list): File attachment list with dicts having 'file_name' and 'url' keys.
        site_code (str): Identifier of the site sending the email.
    """
    config = get_braze_configuration(site_code)
    try:
        user_emails = [user_email]
        braze_client = get_braze_client(site_code)
        response = _send_braze_message(
            braze_client,
            email_ids=user_emails,
            subject=subject,
            body=email_body,
            sender_alias=sender_alias,
            reply_to=reply_to,
            attachments=attachments,
            campaign_id=config.get('ENTERPRISE_CODE_ASSIGNMENT_CAMPAIGN_ID'),
            message_variation_id=config.get('ENTERPRISE_CODE_ASSIGNMENT_MESSAGE_VARIATION_ID'),
        )
        if response and response['success']:
            dispatch_id = response['dispatch_id']
            if update_assignment_email_status(offer_assignment_id, dispatch_id, 'success'):
                logger.info('[Offer Assignment] Offer assignment notification sent with message --- '
                            '{message}'.format(message=email_body))
            else:
                logger.exception(
                    '[Offer Assignment] An error occurred while updating email status data for '
                    'offer {token_offer} and email {token_email} via the ecommerce API.'.format(
                        token_offer=offer_assignment_id,
                        token_email=user_email,
                    )
                )
    except (BrazeRateLimitError, BrazeInternalServerError) as exc:
        raise self.retry(countdown=config.get('BRAZE_RETRY_SECONDS'),
                         max_retries=config.get('BRAZE_RETRY_ATTEMPTS')) from exc
    except BrazeError:
        logger.exception(
            '[Offer Assignment] Error in offer assignment notification with message --- '
            '{message}'.format(message=email_body)
        )


def send_offer_update_email_via_braze(self, user_email, subject, email_body, sender_alias, reply_to, attachments,
                                      site_code):
    """
    Sends the offer emails after assignment via braze, either for revoking or reminding.

    Args:
        self: Ignore.
        user_email (str): Recipient's email address.
        subject (str): Email subject.
        email_body (str): The body of the email.
        site_code (str): Identifier of the site sending the email.
        sender_alias (str): Enterprise Customer sender alias used as From Name.
        reply_to (str): Enterprise Customer reply to address for email reply.
        attachments (list): File attachment list with dicts having 'file_name' and 'url' keys.
    """
    config = get_braze_configuration(site_code)
    try:
        user_emails = [user_email]
        braze_client = get_braze_client(site_code)
        _send_braze_message(
            braze_client,
            email_ids=user_emails,
            subject=subject,
            body=email_body,
            sender_alias=sender_alias,
            reply_to=reply_to,
            attachments=attachments,
            campaign_id=config.get('ENTERPRISE_CODE_UPDATE_CAMPAIGN_ID'),
            message_variation_id=config.get('ENTERPRISE_CODE_UPDATE_MESSAGE_VARIATION_ID'),
        )
    except (BrazeRateLimitError, BrazeInternalServerError) as exc:
        raise self.retry(countdown=config.get('BRAZE_RETRY_SECONDS'),
                         max_retries=config.get('BRAZE_RETRY_ATTEMPTS')) from exc
    except BrazeError:
        logger.exception(
            '[Offer Assignment] Error in offer update notification with message --- '
            '{message}'.format(message=email_body)
        )


def send_offer_usage_email_via_braze(self, emails, subject, email_body, reply_to, attachments, site_code):
    """
    DEPRECATED: This function will eventually be removed in favor of
    ``send_offer_usage_email_via_braze_api_triggered_campaign`` below.
    See https://2u-internal.atlassian.net/browse/ENT-5940

    Sends the offer usage email via braze.
    Args:
        self: Ignore.
        emails (str): comma separated emails.
        subject (str): Email subject.
        email_body (str): The body of the email.
        reply_to (str): Enterprise Customer reply to address for email reply.
        attachments (list): File attachment list with dicts having 'file_name' and 'url' keys.
        site_code (str): Identifier of the site sending the email.
    """
    config = get_braze_configuration(site_code)
    try:
        user_emails = list(emails.strip().split(","))
        braze_client = get_braze_client(site_code)
        _send_braze_message(
            braze_client,
            email_ids=user_emails,
            subject=subject,
            body=email_body,
            reply_to=reply_to,
            attachments=attachments,
            campaign_id=config.get('ENTERPRISE_CODE_USAGE_CAMPAIGN_ID'),
            message_variation_id=config.get('ENTERPRISE_CODE_USAGE_MESSAGE_VARIATION_ID'),
        )
    except (BrazeRateLimitError, BrazeInternalServerError) as exc:
        raise self.retry(countdown=config.get('BRAZE_RETRY_SECONDS'),
                         max_retries=config.get('BRAZE_RETRY_ATTEMPTS')) from exc
    except BrazeError:
        logger.exception(
            '[Offer Usage] Error in offer usage notification with message --- '
            '{message}'.format(message=email_body)
        )


def send_api_triggered_offer_usage_email_via_braze(
    self, lms_user_ids_by_email, subject, email_body_variables, site_code, campaign_id=None
):
    """
    Sends the offer usage email via braze.

    Args:
        self: Ignore.
        lms_user_ids_by_email (dict): Map of recipient email addresses to LMS user ids.
        subject (str): Email subject.
        email_body_variables (dict): key-value pairs that are injected into Braze email template for personalization.
        site_code (str): Identifier of the site sending the email.
        campaign_id (str): Identifier of Braze API-triggered campaign to send message through; defaults
            to config.ENTERPRISE_CODE_USAGE_CAMPAIGN_ID
    """
    config = get_braze_configuration(site_code)
    try:
        braze_client = EdxBrazeClient(site_code)

        message_kwargs = {
            'campaign_id': campaign_id or config.get('ENTERPRISE_CODE_USAGE_API_TRIGGERED_CAMPAIGN_ID'),
            'recipients': [],
            'emails': [],
            'trigger_properties': {
                'subject': subject,
                **email_body_variables,
            },
        }

        for user_email, lms_user_id in sorted(lms_user_ids_by_email.items(), key=itemgetter(0)):
            if lms_user_id:
                recipient = braze_client.create_recipient(user_email, lms_user_id)
                message_kwargs['recipients'].append(recipient)
            else:
                message_kwargs['emails'].append(user_email)

        braze_client.send_campaign_message(**message_kwargs)
    except (edx_braze_exceptions.BrazeRateLimitError, edx_braze_exceptions.BrazeInternalServerError) as exc:
        raise self.retry(
            countdown=OFFER_USAGE_RETRY_DELAY_SECONDS,
            max_retries=config.get('BRAZE_RETRY_ATTEMPTS')
        ) from exc
    except edx_braze_exceptions.BrazeError:
        logger.exception(
            '[Offer Usage] Error in offer usage notification with message --- '
            '{message}'.format(message=email_body_variables)
        )
        raise


def send_code_assignment_nudge_email_via_braze(self, email, subject, email_body, sender_alias, reply_to,  # pylint: disable=invalid-name
                                               attachments, site_code):
    """
    Sends the code assignment nudge email via braze.

    Args:
        self: Ignore.
        email (str): Recipient's email address.
        subject (str): Email subject.
        email_body (str): The body of the email.
        sender_alias (str): Enterprise Customer sender alias used as From Name.
        reply_to (str): Enterprise Customer reply to address for email reply.
        attachments (list): File attachment list with dicts having 'file_name' and 'url' keys.
        site_code (str): Identifier of the site sending the email.
    """
    config = get_braze_configuration(site_code)
    try:
        user_emails = [email]
        braze_client = get_braze_client(site_code)
        _send_braze_message(
            braze_client,
            email_ids=user_emails,
            subject=subject,
            body=email_body,
            sender_alias=sender_alias,
            reply_to=reply_to,
            attachments=attachments,
            campaign_id=config.get('ENTERPRISE_CODE_NUDGE_CAMPAIGN_ID'),
            message_variation_id=config.get('ENTERPRISE_CODE_NUDGE_MESSAGE_VARIATION_ID'),
        )
    except (BrazeRateLimitError, BrazeInternalServerError) as exc:
        raise self.retry(countdown=config.get('BRAZE_RETRY_SECONDS'),
                         max_retries=config.get('BRAZE_RETRY_ATTEMPTS')) from exc
    except BrazeError:
        logger.exception(
            '[Code Assignment Nudge Email] Error in offer nudge notification with message --- '
            '{message}'.format(message=email_body)
        )


def _send_braze_message(braze_client, **kwargs):
    """
    Helper to send braze messages.  Pops any falsey `campaign_id`
    argument before sending.
    """
    if 'campaign_id' in kwargs and not kwargs['campaign_id']:
        kwargs.pop('campaign_id')
    return braze_client.send_message(**kwargs)
